#include <proto.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>

#include "gttextarea.h"
#include "gtapp.h"
#include "gterror.h"
#include "gtresname.h"
#include "gtcommon.h"

/************************************************************************\
*
* GtTextArea class.
*
* $Ident: This file is part of the 'Gt-library' 1.0.0 $
* $Copyright: (C) 1997 by Gran Gustafsson $
*
\************************************************************************/

#define GTTAA_STRING  (GTDSA_MAX_NR)
#define GTTAA_ATOM_NR (GTDSA_MAX_NR+1)

static Atom selatoms[GTTAA_ATOM_NR];

/************************************************************************\
*
* Create and init textarea.
*
\************************************************************************/
GtTextArea *GtMkTextArea(char *n,GtManager *p,GtTag *tgs)
{
	GtTextArea *obj;
	
	if(!(obj=new GtTextArea()))
	{
		GtError(NULL,GTEC_NOOBJECT,"No GtTextArea object!\n");
		return(NULL);
	}
	if(!obj->Init(n,p,tgs))
	{
		GtDestroy(obj);
		obj=NULL;
	}
	return(obj);
}
	
/************************************************************************\
*
* Create and init textarea.
*
\************************************************************************/
GtTextArea *GtMkTextArea(char *n,GtManager *p,ulong tag,...)
{
	va_list va;
	GtTextArea *obj;
	int ps;
	
	va_start(va,tag);
	ps=GtBeginVaTags();
	obj=GtMkTextArea(n,p,GtCopyVaTags(tag,&va));
	GtEndVaTags(ps);
	va_end(va);
	return(obj);
}
	
/************************************************************************\
*
* Setup empty textarea.
*
\************************************************************************/
GtTextArea::GtTextArea(void)
{
	txt=NULL;
	attr=NULL;
	cattr=GTTA_PLAIN;
	tw=1;
	th=1;
	trows=0;
	tcols=0;
	crows=0;
	ccols=0;
	mrows=0;
	mcols=0;
	sr1=0;
	sr2=-1;
	ftype=GTF_IN;
	bdrw=1;
	ibdrw=2;
	press=FALSE;
}

/************************************************************************\
*
* Destroy textarea.
*
\************************************************************************/
void GtTextArea::Destroy(void)
{
	int n;
	
	if(txt)
	{
		for(n=0; n<crows; n++)
		{
			if(txt[n])
				free(txt[n]);
		}
		free(txt);
	}
	if(attr)
	{
		for(n=0; n<crows; n++)
		{
			if(attr[n])
				free(attr[n]);
		}
		free(attr);
	}
	GtWidget::Destroy();
}

/************************************************************************\
*
* Setup textarea.
*
\************************************************************************/
sbool GtTextArea::Init(char *n,GtManager *p,GtTag *tgs)
{
	GtTag *t;
	int mxs,mys,mxxs,mxys;
	XCharStruct xcs;
	
	if(!GtWidget::Init(n,p,tgs))
		return(FALSE);
	app->GetRes(this,GTR_FRAME,GTRT_FRAME,&ftype);
	app->GetRes(this,GTR_BDR,GTRT_BYTE,&bdrw);
	app->GetRes(this,GTR_IBDR,GTRT_BYTE,&ibdrw);
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_Columns:
				tcols=t->d.val;
				break;
			case GT_Rows:
				trows=t->d.val;
				break;
			case GT_TextAttr:
				cattr=t->d.val;
				break;
			case GT_FrameType:
				ftype=t->d.val;
				if(ftype>GTF_LAST_NR)
					ftype=GTF_IN;
				break;
			case GT_Border:
				bdrw=t->d.val;
				break;
			case GT_InnerBorder:
				ibdrw=t->d.val;
				break;
		}
	}
	memcpy(selatoms,dsp->GetStdAtoms(),GTDSA_MAX_NR*sizeof(Atom));
	selatoms[GTTAA_STRING]=XA_STRING;                                
  this->PrivateGC(GT_Foreground);
	switch(ftype)
	{
		case GTF_NONE:
			ibw=ibdrw;
			break;
		case GTF_SIMPLE:
		case GTF_IN:
		case GTF_OUT:
			ibw=bdrw+ibdrw;
			break;
		case GTF_DBL_IN:
		case GTF_DBL_OUT:
			ibw=2*bdrw+ibdrw;
			break;
	}
	gtfnt->MaxExtents(&xcs);
	tw=xcs.width;
	toff=xcs.ascent;
	th=toff+xcs.descent;
	if(tcols)
	{
		mxs=tw*tcols;
		mxxs=tw*tcols;
	}
	else
	{
		mxs=tw;
		mxxs=GTS_MAX;
	}
	if(trows)
	{
		mys=th*trows;
		mxys=th*trows;
	}
	else
	{
		mys=th;
		mxys=GTS_MAX;
	}
	this->SetMinMaxSize(2*ibw+mxs,2*ibw+mys,2*ibw+mxxs,2*ibw+mxys);
	return(TRUE);
}

/************************************************************************\
*
* Set textarea attributes.
*
\************************************************************************/
sbool GtTextArea::Set(GtTag *tgs)
{
	GtTag *t;
	int mxs,mys,mxxs,mxys;
	sbool ch=FALSE,vch=FALSE,fch=FALSE;
	XCharStruct xcs;
	
	if(!GtWidget::Set(tgs))
		return(FALSE);
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_Columns:
				tcols=t->d.val;
				ch=TRUE;
				break;
			case GT_Rows:
				trows=t->d.val;
				ch=TRUE;
				break;
			case GT_TextAttr:
				cattr=t->d.val;
				break;
			case GT_FrameType:
				ftype=t->d.val;
				if(ftype>GTF_LAST_NR)
					ftype=GTF_IN;
				vch=TRUE;
				break;
			case GT_Border:
				bdrw=t->d.val;
				vch=TRUE;
				break;
			case GT_InnerBorder:
				ibdrw=t->d.val;
				vch=TRUE;
				break;
			case GT_Font:
			case GT_FontList:
				fch=TRUE;
				break;
		}
	}
	if(vch)
	{
		switch(ftype)
		{
			case GTF_NONE:
				ibw=ibdrw;
				break;
			case GTF_SIMPLE:
			case GTF_IN:
			case GTF_OUT:
				ibw=bdrw+ibdrw;
				break;
			case GTF_DBL_IN:
			case GTF_DBL_OUT:
				ibw=2*bdrw+ibdrw;
				break;
		}
	}
	if(fch)
	{
		gtfnt->MaxExtents(&xcs);
		tw=xcs.width;
		toff=xcs.ascent;
		th=toff+xcs.descent;
	}
	if(tcols)
	{
		mxs=tw*tcols;
		mxxs=tw*tcols;
	}
	else
	{
		mxs=tw;
		mxxs=GTS_MAX;
	}
	if(trows)
	{
		mys=th*trows;
		mxys=th*trows;
	}
	else
	{
		mys=th;
		mxys=GTS_MAX;
	}
	if(ch || fch || vch)
	{
		this->SetMinMaxSize(2*ibw+mxs,2*ibw+mys,2*ibw+mxxs,2*ibw+mxys);
		this->Refresh();
	}
	return(TRUE);
}
	
	
/************************************************************************\
*
* Get textarea attributes.
*
\************************************************************************/
void GtTextArea::Get(GtTag *tgs)
{
	GtTag *t;
	long *ptr;
	
	GtWidget::Get(tgs);
	while((t=GtNextTag(&tgs)))
	{
		if(!(ptr=(long *)t->d.ptr))
			continue;
		switch(t->tag)
		{
			case GT_Columns:
				*ptr=ccols;
				break;
			case GT_Rows:
				*ptr=crows;
				break;
			case GT_TextAttr:
				*ptr=cattr;
				break;
			case GT_FrameType:
				*ptr=ftype;
				break;
			case GT_Border:
				*ptr=bdrw;
				break;
			case GT_InnerBorder:
				*ptr=ibdrw;
				break;
		}
	}
}

/************************************************************************\
*
* Check if class is (sub)class of named class.
*
\************************************************************************/
sbool GtTextArea::IsSubClassOf(char *cn)
{
	if(strcmp(cn,"GtTextArea")==0)
		return(TRUE);
	else
		return(GtWidget::IsSubClassOf(cn));
}

/************************************************************************\
*
* Refresh part of the textarea.
*
\************************************************************************/
void GtTextArea::RefreshRect(int rxp,int ryp,int rxs,int rys)
{
	int xp,yp,x,y,l;
	uword a;
	
	if(!gfx || (rxp>=ccols) || (rxp+rxs<0) || (ryp>=crows) || (ryp+rys<0))
		return;
	if(rxp+rxs>ccols)
		rxs=ccols-rxp;
	if(ryp+rys>crows)
		rys=crows-ryp;
	gtwin->Clear(ibw+rxp*tw,ibw+ryp*th,rxs*tw,rys*th);
	yp=ibw+ryp*th;
	for(y=0; y<rys; y++)
	{
		xp=ibw+rxp*tw;
		l=0;
		a=attr[ryp+y][rxp];
		for(x=0; x<rxs; x++)
		{
			if(attr[ryp+y][rxp+x]!=a)
			{
				DrawLinePart(xp,yp,ryp+y,rxp+l,rxp+x-1,a);
				xp+=(x-l)*tw;
				l=x;
				a=attr[ryp+y][rxp+x];
			}
		}
		DrawLinePart(xp,yp,ryp+y,rxp+l,rxp+x-1,a);
		yp+=th;
	}
}

/************************************************************************\
*
* Set textarea event mask.
*
\************************************************************************/
void GtTextArea::SetEMask(ulong em)
{
	GtWidget::SetEMask(em|ExposureMask|ButtonPressMask|ButtonReleaseMask|
												Button1MotionMask);
}

/************************************************************************\
*
* Put character on textarea.
*
\************************************************************************/
void GtTextArea::PutCh(int xp,int yp,ubyte c)
{
	if(!txt || (xp<0) || (xp>=ccols) || (yp<0) || (yp>=crows))
		return;
	txt[yp][xp]=c;
	attr[yp][xp]=cattr;
	this->RefreshRect(xp,yp,1,1);
}

/************************************************************************\
*
* Put character and attribute on textarea.
*
\************************************************************************/
void GtTextArea::PutCh(int xp,int yp,ubyte c,uword a)
{
	uword tmp;
	
	tmp=cattr;
	cattr=a;
	PutCh(xp,yp,c);
	cattr=tmp;
}

/************************************************************************\
*
* Get character from textarea.
*
\************************************************************************/
void GtTextArea::GetCh(int xp,int yp,ubyte *c)
{
	if(!txt || (xp<0) || (xp>=ccols) || (yp<0) || (yp>=crows))
		return;
	*c=txt[yp][xp];
}

/************************************************************************\
*
* Get character and attribute from textarea.
*
\************************************************************************/
void GtTextArea::GetCh(int xp,int yp,ubyte *c,uword *a)
{
	if(!txt || (xp<0) || (xp>=ccols) || (yp<0) || (yp>=crows))
		return;
	*c=txt[yp][xp];
	*a=attr[yp][xp];
}

/************************************************************************\
*
* Put string on textarea.
*
\************************************************************************/
int GtTextArea::PutStr(int xp,int yp,char *str,int len)
{
	ubyte *s;
	int n;
	
	if(!txt || (xp>=ccols) || (xp+len<0) || (yp<0) || (yp>=crows))
		return(0);
	if(len<0)
		len=strlen(str);
	s=(ubyte *)str;
	if(xp<0)
	{
		s+=-xp;
		len+=xp;
		xp=0;
	}
	if(xp+len>ccols)
		len=ccols-xp;
	for(n=0; n<len; n++)
	{
		txt[yp][xp+n]=s[n];
		attr[yp][xp+n]=cattr;
	}
	this->RefreshRect(xp,yp,len,1);
	return(len);
}

/************************************************************************\
*
* Put string and attribute on textarea.
*
\************************************************************************/
int GtTextArea::PutStr(int xp,int yp,char *str,int len,uword a)
{
	uword tmp;
	int ret;
	
	tmp=cattr;
	cattr=a;
	ret=PutStr(xp,yp,str,len);
	cattr=tmp;
	return(ret);
}

/************************************************************************\
*
* Fill rectangle of textarea with a character.
*
\************************************************************************/
void GtTextArea::Fill(int xp,int yp,int xs,int ys,ubyte c)
{
	int x,y;
	
	if(!txt || (xp>=ccols) || (xp+xs<0) || (yp>=crows) || (yp+ys<0))
		return;
	if(xp<0)
	{
		xs+=xp;
		xp=0;
	}
	if(yp<0)
	{
		ys+=yp;
		yp=0;
	}
	if(xp+xs>ccols)
		xs=ccols-xp;
	if(yp+ys>crows)
		ys=crows-yp;
	for(y=0; y<ys; y++)
	{
		for(x=0; x<xs; x++)
		{
			txt[yp+y][xp+x]=c;
			attr[yp+y][xp+x]=cattr;
		}
	}
	this->RefreshRect(xp,yp,xs,ys);
}

/************************************************************************\
*
* Fill rectangle of textarea with a character and attribute.
*
\************************************************************************/
void GtTextArea::Fill(int xp,int yp,int xs,int ys,ubyte c,uword a)
{
	uword tmp;
	
	tmp=cattr;
	cattr=a;
	Fill(xp,yp,xs,ys,c);
	cattr=tmp;
}

/************************************************************************\
*
* Set scroll region.
*
\************************************************************************/
void GtTextArea::ScrollRegion(int s1,int s2)
{
	if(s1>crows)
		s1=crows-1;
	if(s2>crows)
		s2=crows-1;
	if(s1<0)
		s1=0;
	if(s2<s1 && (s2>=0))
		s2=s1;
	sr1=s1;
	sr2=s2;
}

/************************************************************************\
*
* Scroll region up or down.
*
\************************************************************************/
void GtTextArea::Scroll(sbool up)
{
	ubyte *ttxt;
	uword *tattr;
	int s1,s2,n,i;
	
	s1=sr1;
	s2=(sr2<0 ? crows-1 : sr2);
	if(s1==s2)
	{
		ttxt=txt[s1];
		tattr=attr[s1];
	}
	else if(up)
	{
		ttxt=txt[s1];
		tattr=attr[s1];
		for(n=s1; n<s2; n++)
		{
			txt[n]=txt[n+1];
			attr[n]=attr[n+1];
		}
		txt[s2]=ttxt;
		attr[s2]=tattr;
	}
	else
	{
		ttxt=txt[s2];
		tattr=attr[s2];
		for(n=s2; n>s1; n--)
		{
			txt[n]=txt[n-1];
			attr[n]=attr[n-1];
		}
		txt[s1]=ttxt;
		attr[s1]=tattr;
	}
	memset(ttxt,' ',sizeof(ubyte)*ccols);
	for(i=0; i<ccols; i++)
		tattr[i]=cattr;
	this->RefreshRect(0,s1,ccols,s2-s1+1);
}

/************************************************************************\
*
* Insert empty line at position.
*
\************************************************************************/
void GtTextArea::InsertLine(int pos)
{
	word os1,os2;
	
	os1=sr1;
	os2=sr2;
	sr2=(sr2<0 ? crows-1 : sr2);
	if(sr1+pos>sr2)
		Scroll(TRUE);
	else
	{
		sr1+=pos;
		Scroll(FALSE);
	}
	sr1=os1;
	sr2=os2;
}

/************************************************************************\
*
* Delete line from textarea at position.
*
\************************************************************************/
void GtTextArea::DeleteLine(int pos)
{
	word os1,os2;
	
	os1=sr1;
	os2=sr2;
	sr2=(sr2<0 ? crows-1 : sr2);
	if(sr1+pos<=sr2)
	{
		sr1+=pos;
		Scroll(TRUE);
	}
	sr1=os1;
	sr2=os2;
}

/************************************************************************\
*
* Set textarea position and size and resize memory as needed.
*
\************************************************************************/
void GtTextArea::SetPosSize(word xp,word yp,word xs,word ys)
{
	int n,i;
	word oxs,oys,ocs,ors;
	
	oxs=xsize;
	oys=ysize;
	GtWidget::SetPosSize(xp,yp,xs,ys);
	if((oxs!=xsize) || (oys!=ysize))
	{
		rect.x=ibw;
		rect.y=ibw;
		rect.width=xsize-2*ibw;
		rect.height=ysize-2*ibw;
		if(gfx)
			gtfore->SetClipRectangles(0,0,&rect,1,YXBanded);
		ocs=ccols;
		ors=crows;
		ccols=(xsize-2*ibw)/tw;
		crows=(ysize-2*ibw)/th;
		if((ccols>mcols) || (crows>mrows))
		{
			oxs=mcols;
			oys=mrows;
			mcols=GTMAX(ccols,mcols);
			mrows=GTMAX(crows,mrows);
			if(!(txt=(ubyte **)realloc(txt,sizeof(ubyte *)*mrows)))
			{
				GtError(this,GTEC_NOMEMORY,"No memory for text!\n");
				return;
			}
			if(!(attr=(uword **)realloc(attr,sizeof(uword *)*mrows)))
			{
				GtError(this,GTEC_NOMEMORY,"No memory for text attributes!\n");
				return;
			}
			for(n=0; n<mrows; n++)
			{
				if(n>=oys)
				{
					txt[n]=NULL;
					attr[n]=NULL;
				}
				if(!txt[n] || (mcols>oxs))
				{
					if(!(txt[n]=(ubyte *)realloc(txt[n],sizeof(ubyte)*mcols)))
					{
						GtError(this,GTEC_NOMEMORY,"No memory for text!\n");
						return;
					}
					if(!(attr[n]=(uword *)realloc(attr[n],sizeof(uword)*mcols)))
					{
						GtError(this,GTEC_NOMEMORY,"No memory for text attributes!\n");
						return;
					}
				}
			}
		}
		if((ccols>=ocs) || (crows>=ors))
		{
			for(n=0; n<crows; n++)
			{
				if(n>=ors)
				{
					memset(txt[n],' ',sizeof(ubyte)*ccols);
					for(i=0; i<ccols; i++)
						attr[n][i]=cattr;
				}
				else if(ccols>ocs)
				{
					memset(&txt[n][ocs],' ',sizeof(ubyte)*(ccols-ocs));
					for(i=ocs; i<ccols; i++)
						attr[n][i]=cattr;
				}
			}
		}
	}
}

/************************************************************************\
*
* Draw textarea when expose.
*
\************************************************************************/
sbool GtTextArea::Draw(XEvent *xe)
{
	int n,x,y,l,xp,yp,rxp,ryp,rxs,rys;
	uword a;
	
	if(!txt || (xe && (xe->type!=Expose)))
		return(TRUE);
	if(xe)
	{
		rxp=(xe->xexpose.x-ibw)/tw;
		ryp=(xe->xexpose.y-ibw)/th;
		rxs=(xe->xexpose.x+xe->xexpose.width-ibw)/tw-rxp+1;
		rys=(xe->xexpose.y+xe->xexpose.height-ibw)/th-ryp+1;
	}
	else
	{
		rxp=0;
		ryp=0;
		rxs=ccols;
		rys=crows;
	}
	if(rxp+rxs>=ccols)
		rxs=ccols-rxp;
	if(ryp+rys>=crows)
		rys=crows-ryp;
	yp=ibw+ryp*th;
	for(y=0; y<rys; y++)
	{
		xp=ibw+rxp*tw;
		l=0;
		a=attr[ryp+y][rxp];
		for(x=0; x<rxs; x++)
		{
			if(attr[ryp+y][rxp+x]!=a)
			{
				DrawLinePart(xp,yp,ryp+y,rxp+l,rxp+x-1,a);
				xp+=(x-l)*tw;
				l=x;
				a=attr[ryp+y][rxp+x];
			}
		}
		DrawLinePart(xp,yp,ryp+y,rxp+l,rxp+x-1,a);
		yp+=th;
	}
	switch(ftype)
	{
		case GTF_SIMPLE:
			for(n=0; n<bdrw; n++)
				gtfore->DrawRectangle(gtwin,n,n,xsize-2*n-1,ysize-2*n-1,FALSE);
			break;
		case GTF_IN:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtdrk,gtlht,gtwin,n,n,xsize-2*n,ysize-2*n);
			}
			break;
		case GTF_OUT:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtlht,gtdrk,gtwin,n,n,xsize-2*n,ysize-2*n);
			}
			break;
		case GTF_DBL_IN:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtdrk,gtlht,gtwin,n,n,xsize-2*n,ysize-2*n);
				GT3DBOX(gtlht,gtdrk,gtwin,bdrw+n,bdrw+n,
								xsize-2*(bdrw+n),ysize-2*(bdrw+n));
			}
			break;
		case GTF_DBL_OUT:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtlht,gtdrk,gtwin,n,n,xsize-2*n,ysize-2*n);
				GT3DBOX(gtdrk,gtlht,gtwin,bdrw+n,bdrw+n,
								xsize-2*(bdrw+n),ysize-2*(bdrw+n));
			}
			break;
	}	
	return(TRUE);
}

/************************************************************************\
*
* Handle textarea mouse events.
*
\************************************************************************/
sbool GtTextArea::Mouse(XEvent *xe)
{
	int xp,yp,osyp,oeyp;
	
	switch(xe->type)
	{
		case ButtonPress:
			if(xe->xbutton.button==Button1)
			{
				xp=xe->xbutton.x;
				yp=xe->xbutton.y;
				if((xp<ibw) || (xp>xsize-ibw-1) || (yp<ibw) || (yp>ysize-ibw-1))
					return(TRUE);
				mxp=(xp-ibw)/tw;
				myp=(yp-ibw)/th;
				msxp=mxp;
				msyp=myp;
				mexp=mxp;
				meyp=myp;
				Mark(TRUE);
				this->RefreshRect(0,msyp,ccols,meyp-msyp+1);
				press=TRUE;
			}
			break;
		case ButtonRelease:
			if(press && (xe->xbutton.button==Button1))
			{
				press=FALSE;
				stime=xe->xbutton.time;
				AllocMark();
				Mark(FALSE);
				this->RefreshRect(0,msyp,ccols,meyp-msyp+1);
			}
			break;
		case MotionNotify:
			if(press)
			{
				xp=xe->xbutton.x;
				yp=xe->xbutton.y;
				if((xp<ibw) || (xp>xsize-ibw-1) || (yp<ibw) || (yp>ysize-ibw-1))
					return(TRUE);
				osyp=msyp;
				oeyp=meyp;
				Mark(FALSE);
				xp=(xp-ibw)/tw;
				yp=(yp-ibw)/th;
				if(((xp==msxp) && (yp==msyp)) || ((xp==mexp) && (xp==meyp)))
					return(TRUE);
				if((yp<myp) || ((yp==myp) && (xp<mxp)))
				{
					msxp=xp;
					msyp=yp;
				}
				else
				{
					mexp=xp;
					meyp=yp;
				}
				Mark(TRUE);
				this->RefreshRect(0,GTMIN(msyp,osyp),ccols,GTMAX(meyp-msyp,oeyp-osyp)+1);
			}
			break;
	}
	return(TRUE);
}

/************************************************************************\
*
* Draw part of a line with it's attributes.
*
\************************************************************************/
void GtTextArea::DrawLinePart(int xp,int yp,int y,int s,int e,uword a)
{
	if(a & (GTTA_REVERSE|GTTA_MARKED))
	{
		if(a & GTTA_MARKED)
			gtlht->DrawRectangle(gtwin,xp,yp,tw*(e-s+1),th,TRUE);
		else
			gtfore->DrawRectangle(gtwin,xp,yp,tw*(e-s+1),th,TRUE);
		gtfore->SetForeground(gfx->cols[GTBACK]);
		gtfore->SetBackground(gfx->cols[GTFORE]);
	}
	gtfore->DrawString(gtwin,xp,yp+toff,(char *)&txt[y][s],e-s+1,FALSE);
	if(a & GTTA_BOLD)
		gtfore->DrawString(gtwin,xp+1,yp+toff,(char *)&txt[y][s],e-s+1,FALSE);
	if(a & GTTA_UNDERLINE)
		gtfore->DrawLine(gtwin,xp,yp+th-1,xp+tw*(e-s+1)-1,yp+th-1);
	if(a & (GTTA_REVERSE|GTTA_MARKED))
	{
		gtfore->SetForeground(gfx->cols[GTFORE]);
		gtfore->SetBackground(gfx->cols[GTBACK]);
	}
}

/************************************************************************\
*
* Allocate selection of marked area.
*
\************************************************************************/
void GtTextArea::AllocMark(void)
{
	ubyte *mem;
	int x,y,n,len;
	
	if(meyp-msyp==0)
		len=mexp-msxp+1;
	else
		len=(ccols-msxp)+mexp+1+(meyp-msyp-1)*ccols+(meyp-msyp);
	len++;
	if(!(mem=(ubyte *)malloc(len)))
	{
		GtError(this,GTEC_NOMEMORY,"No memory for selection!\n");
		return;
	}
	n=0;
	for(y=msyp; y<=meyp; y++)
	{
		for(x=(y==msyp ? msxp : 0); x<=(y==meyp ? mexp : ccols-1); x++)
		{
			mem[n]=txt[y][x];
			n++;
		}
		mem[n]='\n';
		n++;
	}
	mem[n]=0;
	if(!dsp->OwnSelection(this,XA_PRIMARY,stime,ConvMark,NULL,DoneMark,mem))
		free(mem);	
}
	
/************************************************************************\
*
* Mark area of text.
*
\************************************************************************/
void GtTextArea::Mark(sbool on)
{
	int x,y;
	
	for(y=msyp; y<=meyp; y++)
	{
		for(x=(y==msyp ? msxp : 0); x<=(y==meyp ? mexp : ccols-1); x++)
		{
			if(on)
				attr[y][x]|=GTTA_MARKED;
			else
				attr[y][x]&=~GTTA_MARKED;
		}
	}
}

/************************************************************************\
*
* Convert marked text.
*
\************************************************************************/
int GtTextArea::ConvMark(GtBase *obj,GtCDPtr cd,GtUDPtr ud)
{
	GtTextArea *ta;
	GtSelReqCD *scd;
	ubyte *mem;
	
	obj=obj;
	scd=(GtSelReqCD *)cd;
	ta=(GtTextArea *)scd->w;
	mem=(ubyte *)ud;
	if(scd->target==selatoms[GTDSA_TARGETS])
	{
		scd->target=XA_ATOM;
		scd->fmt=32;
		scd->data=(ubyte *)selatoms;
		scd->len=GTTAA_ATOM_NR;
	}
	else if(scd->target==selatoms[GTTAA_STRING])
	{
		scd->target=selatoms[GTTAA_STRING];
		scd->fmt=8;
		scd->data=mem;
		scd->len=strlen((char *)mem);
	}
	return(GTCB_CONT);
}

/************************************************************************\
*
* Marked text not needed any more.
*
\************************************************************************/
int GtTextArea::DoneMark(GtBase *obj,GtCDPtr cd,GtUDPtr ud)
{
	obj=obj;
	cd=cd;
	free((ubyte *)ud);
	return(GTCB_CONT);
}
